// ParallelEdges.cpp: Entry point.
//

#include "stdafx.h"
#include <vector>
#include <iostream>
#include <string>
#include <fstream>

#include "stdio.h"

#include "Tools_Geometry.h"

#include "CueParallelEdges.h"


int _tmain(int argc, _TCHAR* argv[])
{

	//Input data
	//----------
	int VertexCount= 0;
	std::vector <POINT2D> Vertex;

	int EdgeCount= 0;
	std::vector <int> EdgeHead;
	std::vector <int> EdgeTail;

	std::string FileName;
	std::cout << "\nPlease enter your desired filename \n (extension .Sketch will be added automatically)\n" << std::endl;
	std::cin >> FileName;


	//std::string FileName= "VP01";
	FileName += ".Sketch";
	std::ifstream in (FileName.c_str());

	//Reads Vertices and Edges
	if (in.is_open()){

		std::string line;
		while (getline(in, line)){
			if(line.substr(0, 6)== "Vertex"){
				POINT2D Vertice;

				long BeginX= line.find(':')+1;
				long EndX= line.find(',', BeginX);
				Vertice.x= atof(line.substr(BeginX, EndX-BeginX).c_str());

				long BeginY= EndX+1;
				while(line.substr(BeginY, 1) == " ")
					BeginY++;
				long EndY= line.find(' ', BeginY);
				if(EndY<0)
					EndY= line.size()-1;
				Vertice.y= -atof(line.substr(BeginY, EndY-BeginY).c_str());

				Vertex.push_back(Vertice);
				VertexCount++;
			}
			else if(line.substr(0, 4)== "Line"){
				EdgeTail.push_back(0);
				long BeginV1 = line.find('V')+1;
				long EndV1 = line.find(' ', BeginV1);
				sscanf_s(line.substr(BeginV1, EndV1-BeginV1).c_str(), "%d", &EdgeTail[EdgeCount]);

				EdgeHead.push_back(0);
				long BeginV2 = line.rfind('V')+1;
				long EndV2 = line.size();
				sscanf_s(line.substr(BeginV2, EndV2-BeginV2).c_str(), "%d", &EdgeHead[EdgeCount]);
				EdgeCount++;
			}
		} //while

		// Closes the file
		in.close();
	}
	else
		printf("Unable to open file\n");

	//Calculate parallel edges
	//------------------------
	CCueParallelEdges PE(VertexCount, Vertex,
						 EdgeCount, EdgeHead, EdgeTail);
	long Approach= 0; //Automatic= 0, Vanishing points= 1, Angular distribution= 2
	std::vector <std::vector <long>> ParallelEdges= PE.GetParallelEdges(Approach);

	//Print output
	//------------
	printf("Example %s \n",FileName.c_str());

	long NumParallelGroups= ParallelEdges.size();
	printf("Number of parallel groups= %3d \n", NumParallelGroups);

	for(long i= 0; i<NumParallelGroups; i++){
		printf("Group number %3d \n", i);
		for(long j= 0; j<(long)ParallelEdges[i].size(); j++){
			printf("Edge %3d \n", ParallelEdges[i][j]);
		}
	}

	printf("\n");
	system("PAUSE");

	return 0;
}

