//+-----------------------------------------------------------------------------+
//| Copyright (C) 2013 REGEO (http://www.regeo.uji.es/)
//|
//| This program is free software
//| You can redistribute it and/or modify it under the terms of the GNU General Public License,
//| as published by the Free Software Foundation.
//| 
//| This program is distributed in the hope that it will be useful,
//| but WITHOUT ANY WARRANTY; without even the implied warranty of
//| MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//+----------------------------------------------------------------------------*/

#include <stdafx.h>
#include <cmath>
#include <algorithm>
#include <cassert>

#include "Tools_Geometry.h"


POINT2D::POINT2D()
:x (0),
 y (0)
{
}


POINT2D::POINT2D(const POINT2D& DP)
:x (DP.x),
 y (DP.y)
{
}


POINT2D::POINT2D(double cx, double cy)
:x (cx),
 y (cy)
{
}


POINT2D::~POINT2D()
{
}


bool POINT2D::operator==(const POINT2D& right) const
//Overload of == operator for POINT2D
{
	return (this->x==right.x && this->y==right.y);
}


bool POINT2D::operator!=(const POINT2D& right) const
//Overload of != operator for POINT2D
{
	return (this->x!=right.x || this->y!=right.y);
}


POINT2D& POINT2D::operator+=(const POINT2D& right)
//Overload of += operator for POINT2D
{
	this->x+= right.x;
	this->y+= right.y;
	return *this;
}


POINT2D POINT2D::operator+(const POINT2D& right) const
//Overload of + operator for POINT2D
{
	POINT2D p(this->x+right.x, this->y+right.y);
	return p;
	//return POINT2D (*this) += right;
}


POINT2D& POINT2D::operator-=(const POINT2D& right)
//Overload of -= operator for POINT2D
{
	this->x-= right.x;
	this->y-= right.y;
	return *this;
}


POINT2D POINT2D::operator-(const POINT2D& right) const
//Overload of - operator for POINT2D
{
	POINT2D p(this->x-right.x, this->y-right.y);
	return p;
	//return POINT2D (*this) -= right;
}


POINT2D& POINT2D::operator*=(const POINT2D& right)
//Overload of *= operator for POINT2D
{
	this->x*= right.x;
	this->y*= right.y;
	return *this;
}


POINT2D POINT2D::operator*(const POINT2D& right) const
//Overload of * operator for POINT2D
{
	POINT2D p(this->x*right.x, this->y*right.y);
	return p;
	//return POINT2D (*this) *= right;
}


POINT2D& POINT2D::operator/=(const POINT2D& right)
//Overload of /= operator for POINT2D
{
	assert(right.x!=0);
	assert(right.y!=0);
	this->x/= right.x;
	this->y/= right.y;
	return *this;
}


POINT2D POINT2D::operator/(const POINT2D& right) const
//Overload of / operator for POINT2D
{
	assert(right.x!=0);
	assert(right.y!=0);
	POINT2D p(this->x/right.x, this->y/right.y);
	return p;
	//return POINT2D (*this) /= right;
}


POINT2D POINT2D::Rotate(POINT2D Center, double beta)
//Returns coordinates of the point rotated and angle "beta" around the point "center"
{
	POINT2D PRot;
	PRot.x= Center.x+((x-Center.x)*cos(beta)-(y-Center.y)*sin(beta));
	PRot.y= Center.y+((x-Center.x)*sin(beta)+(y-Center.y)*cos(beta));
	return PRot;
}


POINT2D POINT2D::Rotate(double beta)
//Returns coordinates of the point rotated and angle "beta" around the origin
{
	POINT2D PRot;
	PRot.x= x*cos(beta)-y*sin(beta);
	PRot.y= x*sin(beta)+y*cos(beta);
	return PRot;
}


SEGMENT2D::SEGMENT2D()
:p0 (0,0),
 p1 (0,0)
{
}


SEGMENT2D::SEGMENT2D(const SEGMENT2D& DS)
:p0 (DS.p0),
 p1 (DS.p1)
{
}


SEGMENT2D::SEGMENT2D(POINT2D cp0, POINT2D cp1)
:p0 (cp0),
 p1 (cp1)
{
}


SEGMENT2D::~SEGMENT2D()
{
}


void SEGMENT2D::Rotate(POINT2D Center, double beta)
//Returns coordinates of the point rotated and angle "beta" around the point "center"
{
	p0= p0.Rotate(Center, beta);
	p1= p1.Rotate(Center, beta);
}


bool SEGMENT2D::operator==(const SEGMENT2D& right) const
//Overload of == operator for SEGMENT2D
{
	return (this->p0==right.p0 && this->p1==right.p1);
}


bool SEGMENT2D::operator!=(const SEGMENT2D& right) const
//Overload of != operator for SEGMENT2D
{
	return (this->p0!=right.p0 || this->p1!=right.p1);
}


RECTANGLE2D::RECTANGLE2D()
:left (0),
 right (0),
 bottom (0),
 top (0)
{
}


RECTANGLE2D::RECTANGLE2D(const RECTANGLE2D& DR)
:left (DR.left),
 right (DR.right),
 bottom (DR.bottom),
 top (DR.top)
{
}


RECTANGLE2D::RECTANGLE2D(double cleft, double cright, double cbottom, double ctop)
:left (cleft),
 right (cright),
 bottom (cbottom),
 top (ctop)
{
}


RECTANGLE2D::~RECTANGLE2D()
{
}


double GetAngle_PI(POINT2D P0, POINT2D P1)
//Gets the angle (in radians) between
//	the segment defined by P0 and P1
//	and the X axis

//Since the edge is not oriented,
//returns a value in the range [0, PI)
{
	double length= MODULE2D(P0.x, P0.y, P1.x, P1.y);
	assert(length > 0);

	double Angle;
	if (P0.y>=P1.y)   // [0, 180]
		Angle= acos((P0.x-P1.x)/length);
	else   // (180, 360]
		Angle= acos((P1.x-P0.x)/length);

	if (fabs(Angle-NUMBER_PI)<Accuracy) Angle= 0.0;

	return Angle;
}


double GetAngle_PI(SEGMENT2D S)
//Gets the angle (in radians) between the segment S and the X axis

//Since the edge is not oriented,
//returns a value in the range [0, PI)
{
	return GetAngle_PI(S.p0, S.p1);
}


double GetAngle_2PI(POINT2D T, POINT2D H)
//Gets the angle (in radians) between
//	the segment defined by T and H
//	and the X axis

//The edge is oriented: T is the tail and H is the head,
//Returns a value in the range [0, 2PI)
{
	double length= MODULE2D(T.x, T.y, H.x, H.y);
	assert(length > 0);

	double Angle;
	if (H.y >= T.y){   // [0, 180]
		Angle= acos((H.x-T.x)/length);
	}
	else{   // (180, 360]
		Angle= acos((T.x-H.x)/length);
		Angle+= NUMBER_PI;
	}
	Angle= std::max(Angle, 0.);
	Angle= std::min(Angle, 2.*NUMBER_PI);
	if (fabs(Angle-2*NUMBER_PI)<Accuracy)
		Angle= 0.0;

	return Angle;
}


double GetAngle_2PI(SEGMENT2D S)
//Gets the angle (in radians) between the segment S and the X axis

//The edge is oriented: S.p0 is the tail and S.p1 is the head,
//Returns a value in the range [0, 2PI)
{
	return GetAngle_2PI(S.p0, S.p1);
}


double GetAngle3Points(POINT2D P1, POINT2D P2, POINT2D P3)
//Gets the angle (in radians) between the two segments defined by P1-P2 and P2-P3

//Returns a value in the range [0, PI]
{
	double Angle= fabs(GetAngle_2PI(P2, P1) - GetAngle_2PI(P2, P3));
	if (Angle>NUMBER_PI)
		Angle= 2*NUMBER_PI-Angle;

	return Angle;
}


double GetDistancePointToSegment(POINT2D p0, POINT2D p1, POINT2D V,
								 POINT2D &Intersection)
//Gets the distance from an external point "V" to a segment defined by p0 and p1
//Gets also the intersection of p0-p1 with the minimum distance segment

//Returns the distance and the intersection point
{
	if(MODULE_2D(p0, p1) == 0)
		return MODULE_2D(p0, V);

	//Get the intersection point
	double u= (V.x - p0.x)*(p1.x - p0.x) + (V.y - p0.y)*(p1.y - p0.y);
	u/= ((p1.x - p0.x)*(p1.x - p0.x) + (p1.y - p0.y)*(p1.y - p0.y));
	Intersection.x = p0.x + ((p1.x - p0.x)*u);
	Intersection.y = p0.y + ((p1.y - p0.y)*u);

	//Get the distance from the external point to the intersection point
	return MODULE_2D(Intersection, V);
}


double GetDistancePointToSegment(POINT2D p0, POINT2D p1, POINT2D V)
//Gets the distance from an external point "V" to a segment defined by p0 and p1
{
	if(MODULE_2D(p0, p1) == 0)
		return MODULE_2D(p0, V);

	return fabs(((V.x - p0.x)*(p1.y - p0.y) - (V.y - p0.y)*(p1.x - p0.x)) / MODULE_2D(p0, p1));
}


bool AreCollinear(POINT2D A0, POINT2D A1,
				  POINT2D B0, POINT2D B1,
				  double ThresholdAngle, double ThresholdDistance)
//Returns true if two segments are collinear, withing a "treshold"
//A0 and A1 are the endpoints of the first segment
//B0 and B1 are the endpoints of the second segment

//The ThresholdAngle is the maximum angle difference (in radians)

//The ThresholdDistance is the maximum perpendicular distance
//between the end nodes of the second segment
//and the line defined by the first segment
//(in % of the minimum lenght of the segments)
{
	double Offset= std::min(MODULE_2D(A0,A1), MODULE_2D(B0,B1));
	Offset*= ThresholdDistance; //% of the legth of the shortest segment

	//Check if segments are nearly coincident
	//---------------------------------------
	if((MODULE_2D(A0,B0)<Offset && MODULE_2D(A1,B1)<Offset) ||
	   (MODULE_2D(A0,B1)<Offset && MODULE_2D(A1,B0)<Offset))
		return true; //Lines overlap (are trivially collinear)

	//Check if worst point of one line is too distant from the other
	//--------------------------------------------------------------
	if(GetDistancePointToSegment(A0, A1, B0) > Offset &&
	   GetDistancePointToSegment(A0, A1, B1) > Offset)
	   return false;

	//Check orientation
	//-----------------
	double angleA= GetAngle_PI(A0,A1);
	double angleB= GetAngle_PI(B0,B1);

	//Return true if differences between both angles are below the threshold
	return ((fabs(angleA-angleB) < ThresholdAngle ||
			 fabs(fabs(angleA-angleB)-NUMBER_PI) < ThresholdAngle));
}
